"""Everything to interact with the A28 API."""
import json
import logging
from argparse import Namespace, _SubParsersAction
from enum import Enum
from typing import Any, Callable, Dict, Union

import certifi
import requests
from requests import Response
from requests.auth import AuthBase

from a28 import utils
from a28.config import ConfigFile


logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
log = logging.getLogger(__name__)


# can be updated to StrEnum when we are on Python 3.10
# https://docs.python.org/3.11/howto/enum.html#strenum
class ApiStatus(str, Enum):
    """Possible response of the API status field."""

    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class ApiError(Exception):
    """Custom exception for API errors."""

    pass


def parse_api_errors(response: Dict[str, Any], http_status_code: int) -> Dict[str, Any]:
    """Raise error if API returned errors or invalid data."""
    if response.get("status") == ApiStatus.SUCCESS:
        return response

    if response.get("message"):
        raise ApiError(response.get("message"))

    raise ApiError(http_status_code)


def handle_api_errors(api_call) -> Callable:
    """Handle API error with a decorator."""

    def wrapper(*arg, **kwargs) -> Dict[str, Any]:
        """Process api response.

        check for errors and raise ApiError if any.
        """
        try:
            response = api_call(*arg, **kwargs)
            response_payload = json.loads(response.content)

            return parse_api_errors(response_payload, response.status_code)
        except json.JSONDecodeError as err:
            raise ApiError(f"Unknown Error: {err}")

    return wrapper


class JwtAuth(AuthBase):
    """Attaches JWT Authentication to the given Request object."""

    def __init__(self, token):
        """Create an Auth object for the requests."""
        self.token = token

    def __call__(self, request):
        """Insert the token in the Authorization header."""
        request.headers["Authorization"] = f"Bearer {self.token}"
        return request


class API:
    """Handles all the actions to the API.

    A region must be given to each instance so we can request to the
    correct API (international, china, etc..).
    """

    ENDPOINTS_URL = "https://assets.a28.io/plugin/endpoints.json"
    DEFAULT_ENDPOINT = "international"

    def __init__(self, region: str):
        """Initialize the API."""
        self.region = region
        self.session = requests.Session()
        self.session.verify = certifi.where()

        api_endpoints = self.fetch_endpoints()
        try:
            self.api_endpoint = api_endpoints[self.region]["api"]["endpoint"]
        except KeyError:
            utils.message("invalid endpoint")
            raise

        self.session.hooks["response"].append(self._check_token)

    @handle_api_errors
    def get(self, url: str, **kwargs) -> Response:
        """Send a GET request."""
        url = f"{self.api_endpoint}{url}"
        kwargs.setdefault("allow_redirects", True)
        return self.session.request("GET", url, **kwargs)

    @handle_api_errors
    def post(self, url: str, data=None, json_payload=None, **kwargs) -> Response:
        """Send a POST request."""
        return self.session.request(
            "POST", self.api_endpoint + url, data=data, json=json_payload, **kwargs
        )

    @handle_api_errors
    def put(self, url: str, data=None, **kwargs) -> Response:
        """Send a PUT request."""
        endpoint = f"{self.api_endpoint}{url}"
        return self.session.request("PUT", endpoint, data=data, **kwargs)

    @handle_api_errors
    def patch(self, url: str, data=None, **kwargs) -> Response:
        """Send a PATCH request."""
        endpoint = f"{self.api_endpoint}{url}"
        return self.session.request("PATCH", endpoint, data=data, **kwargs)

    @handle_api_errors
    def delete(self, url: str, **kwargs) -> Response:
        """Send a DELETE request."""
        endpoint = f"{self.api_endpoint}{url}"
        return self.session.request("DELETE", endpoint, **kwargs)

    def _check_token(self, response: Response, *_args, **_kwargs) -> None:
        """
        Update the local token to a token returned by a server.

        When a server responds to a request, the server can return a
        newly updated token. The new token could be in response to new
        information stored within the token, or the token age might be
        getting close to the expiry time and a token refresh is needed.
        For more information on bearer tokens, please refer to
        `bearer tokens <https://datatracker.ietf.org/doc/html/rfc6750>`.

        This method checks if a updated token has been returned and
        stores it in the preferences system for future requests if the
        token is valid.

        This function is called as a requests hook.
        """
        try:
            json_data = response.json()
        except json.JSONDecodeError:
            return

        if "token" not in json_data:
            return

        with ConfigFile() as data:
            if self.region not in data:
                data[self.region] = {}
            data[self.region]["token"] = json_data["token"]

    def fetch_endpoints(self) -> Dict[str, Any]:
        """Fetch the Area28 endpoint list.

        Returns the decoded JSON to the calling method.
        """
        req = self.session.get(self.ENDPOINTS_URL, allow_redirects=True)
        try:
            regions = json.loads(req.content)
            regions["dev"] = {"api": {"endpoint": "https://dev-api.a28.io/v1"}}
            regions["sit"] = {"api": {"endpoint": "https://sit-api.a28.io/v1"}}
            return regions
        except json.JSONDecodeError as error:
            utils.message(f"error decoding endpoints: {error}")
            raise


def get_package(region: str, identifier: str) -> Union[Dict[str, Any], None]:
    """Retrieve a package from the API."""
    response = f'{{"region": {region}, "identifier": {identifier}}}'
    log.debug(f"retrieving package: {response}")

    api = API(region)
    config = ConfigFile.load()

    try:
        response = api.get(
            f"/package/{identifier}", auth=JwtAuth(config[api.region]["token"])
        )

        return response.get("data")
    except ApiError as error:
        utils.message(f"unable to get package: {error}")
        raise


def register_package(
    region: str, scope: str, name: str, schema: str
) -> Union[str, None]:
    """Create a package with the API and returns the new uuid."""
    log.debug(f'creating package: {{"region": {region}, "name": {name}}}')

    api = API(region)
    config = ConfigFile.load()

    try:
        response = api.post(
            f"/group/{scope}/package",
            json_payload={
                "name": name,
                "type": schema,
            },
            auth=JwtAuth(config[api.region]["token"]),
        )

        return response["data"]["identifier"]
    except ApiError as error:
        utils.message(f"unable to create package: {error}")
        raise


def status(args: Namespace) -> None:
    """Display the status of an endpoint.

    Submit an options request to an api endpoint to check if the
    endpoint returns a 200 or a 304 status response message with a JSON
    payload that has a status response and a StatusCode set.
    """
    api = API(args.endpoint)
    try:
        response = api.get("/status")
        utils.message(response["message"])
    except (ApiError, KeyError):
        utils.message("unable to connect to endpoint")
        raise


def endpoints(_args: Namespace) -> None:
    """Fetch and print the endpoints."""
    api = API(API.DEFAULT_ENDPOINT)

    for endpoint in api.fetch_endpoints():
        utils.message(endpoint)


def find_or_create_package(
    region: str,
    schema: Union[str, None] = None,
    scope: Union[str, None] = None,
    name: Union[str, None] = None,
    identifier: Union[str, None] = None,
) -> Union[str, None]:
    """Find or create a package and returns its UUID.

    if an identifier is provided we'll retrieve it from the API.
    Else we'll create it.
    """
    if identifier:
        existing_package = get_package(region, identifier)

        if not existing_package:
            return

        return existing_package.get("identifier")

    return register_package(
        region=region,
        schema=schema,
        scope=scope,
        name=name,
    )


def cli_options(main_parser: _SubParsersAction) -> None:
    """Argument for the API subcommand."""
    parser = main_parser.add_parser("api", help="api actions")
    api_parser = parser.add_subparsers(
        dest="api",
        required=True,
        help="api",
    )
    parser_endpoints = api_parser.add_parser(
        "endpoints",
        help="list all endpoints.",
    )
    parser_endpoints.set_defaults(func=endpoints)
    parser_status = api_parser.add_parser(
        "status",
        help="show the status of an endpoint.",
    )
    parser_status.add_argument(
        "-e",
        "--endpoint",
        required=True,
        help="the name of the endpoint to check the status of.",
    )
    parser_status.set_defaults(func=status)



"""Manage account actions.

Using the requests module, authenticate an account on the API and store
the authentication token within a local file.
"""
from argparse import Namespace, _SubParsersAction
from getpass import getpass

from a28 import utils
from a28.api import API, ApiError
from a28.config import ConfigFile


def cli_options(main_parser: _SubParsersAction) -> None:
    """Account subcommand arguments."""
    parser = main_parser.add_parser("account", aliases=["acc"], help="account actions")
    subparser = parser.add_subparsers(
        dest="account",
        required=True,
        help="account",
    )
    parser_authenticate = subparser.add_parser(
        "authenticate",
        help="authenticate with the endpoint.",
    )
    parser_authenticate.set_defaults(func=authenticate_interactive)
    parser_authenticate.add_argument(
        "-e",
        "--endpoint",
        default=API.DEFAULT_ENDPOINT,
        help="endpoint to use.",
    )
    parser_authenticate.add_argument(
        "-u",
        "--email",
        required=True,
        help="your email address.",
    )
    parser_authenticate.add_argument(
        "-p",
        "--password",
        help="your password.",
    )


def authenticate_interactive(args: Namespace) -> None:
    """Get the user's password."""
    if args.password:
        message = (
            "Warning: Using a password on the command line"
            " interface can be insecure."
        )
        utils.message(message)
        password = args.password
    else:
        password = getpass(prompt="Enter your password:")

    authenticate(args.endpoint, args.email, password)


def authenticate(region: str, email: str, password: str) -> None:
    """Authenticate to the API."""
    api = API(region)

    try:
        api.post(
            url="/account/signin", json_payload={"email": email, "password": password}
        )

        # save the email for reference
        # token will be saved by req hook (see _check_token function)
        with ConfigFile() as data:
            data[api.region]["email"] = email

        utils.message("authentication successful")
    except ApiError as error:
        utils.message(f"error authenticating: {error}")
        raise



"""Manage package creation, building, and submitting.

The package module provides a set of utilities designed for package
developers to create packages, build packages from python source code,
and submit packages to be included in the list of installable packages
within Area28.
"""
import hashlib
import json
import os
import shutil
from argparse import Namespace

from a28 import log, utils
from a28.api import API, find_or_create_package
from a28.build import add_build_parser, extract_meta
from a28.config import ConfigFile
from a28.publish import add_publish_parser

package_types = [
    "api",
    "app",
    "chat",
    "event",
    "interface",
    "logger",
    "metadata",
    "preference",
    "realtime",
    "repository",
    "ui",
    "units",
]


def cli_options(main_parser):
    """Argparse options added to the cli."""
    parser = main_parser.add_parser("package", aliases=["pkg"], help="package actions")
    package_parser = parser.add_subparsers(
        dest="package",
        required=True,
        help="package",
    )

    add_init_parser(
        package_parser.add_parser(
            "init",
            help="Initialize a package",
        )
    )
    add_meta_parser(
        package_parser.add_parser("meta", help="Display package information")
    )
    add_build_parser(
        package_parser.add_parser(
            "build",
            help="build package",
        )
    )
    add_install_parser(
        package_parser.add_parser(
            "install",
            help="install package",
        )
    )
    add_publish_parser(
        package_parser.add_parser(
            "publish",
            help="publish a package",
        )
    )


def add_meta_parser(parser_meta) -> None:
    """Create the meta command."""
    parser_meta.set_defaults(func=meta)
    parser_meta.add_argument("path", default=".", help='package path "."')


def add_init_parser(parser_init):
    """Create the CLI parser for 'init' command."""
    parser_init.set_defaults(func=initialize)
    parser_init.add_argument(
        "path",
        default=".",
        help='package path default "."',
    )
    parser_init.add_argument(
        "-s",
        "--scope",
        required=True,
        help="package scope eg. group-name",
    )
    parser_init.add_argument(
        "-n",
        "--name",
        required=True,
        help="package name eg. pkg-name",
    )
    parser_init.add_argument(
        "-t",
        "--type",
        required=True,
        choices=package_types,
        help="package type",
    )
    parser_init.add_argument(
        "-i",
        "--identifier",
        required=False,
        help="provide a pre-registered identifier (defaults to API value)",
    )
    parser_init.add_argument(
        "--bin",
        action="store_true",
        help="create the bin directory",
    )
    parser_init.add_argument(
        "--script",
        action="store_true",
        help="create the scripts directory",
    )
    parser_init.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force overwriting package.json",
    )
    parser_init.add_argument(
        "-e",
        "--endpoint",
        default=API.DEFAULT_ENDPOINT,
        help="endpoint to use.",
    )


def add_install_parser(parser_install):
    """Create the CLI parser for 'install' command."""
    parser_install.set_defaults(func=install)
    parser_install.add_argument(
        "--pkg",
        required=True,
        help="package a28 package file",
    )


def generate_sub_dirs(args):
    """Generate a list of sub directories from the give args."""
    sub_dirs = ["extensions", "plugins"]

    if args.bin:
        log.debug("adding bin")
        sub_dirs.append("bin")
    if args.script:
        log.debug("adding scripts")
        sub_dirs.append("script")
    if args.type == "application":
        sub_dirs.append("plugin")

    return sub_dirs


def generate_jsondata(scope, name, schema, identifier, args):
    """Generate a basic package.json data."""
    json_data = {
        "name": f"@{scope}/{name}",
        "description": f"{schema} package created using A28 cmd by {scope}.",
        "version": "0.1.0",
        "identifier": identifier,
    }

    if args.bin:
        json_data["bin"] = {}
    if args.script:
        json_data["scripts"] = {}

    return json_data


def meta(args) -> None:
    """Display package meta information."""
    package = os.path.join(args.path, "package.json")

    if not os.path.exists(package):
        utils.message(f"invalid package path {package}")

    info = get_info(args.path)
    for name, value in info.items():
        utils.message(f"{name}: {value}")


def initialize(args: Namespace):
    """Initialize a package in a specified directory."""
    (path, name, scope, schema) = _get_init_args(args)
    log.info(f'initializing @{scope}/{name} in folder "{path}"')

    # check if the user has authenticated
    if not ConfigFile.load():
        utils.message("authenticate before initializing")
        raise SystemExit(1)

    identifier = find_or_create_package(
        region=args.endpoint,
        identifier=args.identifier,
        schema=schema,
        scope=scope,
        name=name,
    )

    if not identifier:
        raise SystemExit(1)

    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    json_file = os.path.join(path, "package.json")

    _build_structure(path, generate_sub_dirs(args))
    overwrite = "overwrite package.json?"
    if not os.path.isfile(json_file) or args.force or utils.confirm(overwrite):
        log.info("creating package.json")
        with open(json_file, "w") as pkg_file:
            json.dump(
                generate_jsondata(scope, name, schema, identifier, args),
                pkg_file,
                indent=4,
            )
    else:
        log.info("not overwriting package.json")
        return

    utils.message("package created")


def install(args):
    """Install / update the local package."""
    pkg_hash = generate_hash(args.pkg)
    meta = extract_meta(args.pkg)
    meta["hash"] = pkg_hash
    install_local(args.pkg, meta)
    utils.message(f'installed {meta["name"]}')


def install_local(pkg, meta):
    """Install package locally by calling the update_index function."""
    dest = os.path.join(utils.STORAGE, "cache")
    shutil.copy(pkg, dest)
    index = os.path.join(utils.STORAGE, "index.json")
    with open(index, "r+") as index_data:
        update_index(index_data, meta)


def update_index(index_data, meta):
    """Update the index.json without touching the etag.json file.

    This will be overwritten when new official index.json is published.
    """
    data = json.load(index_data)
    data["packages"][meta["name"]] = meta
    if meta["name"] not in data["required"]:
        data["required"].append(meta["name"])
    index_data.seek(0)
    json.dump(data, index_data, indent=4)
    index_data.truncate()


def generate_hash(pkg):
    """Generate a hash for a package."""
    sha256_hash = hashlib.sha256()
    with open(pkg, "rb") as f:
        # read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def _get_init_args(args):
    path = os.path.realpath(args.path)
    if not utils.valid_string(args.name):
        raise ValueError(f"{args.name} is not a valid name")
    name = args.name.lower()
    if not utils.valid_string(args.scope):
        raise ValueError(f"{args.scope} is not a valid scope")
    scope = args.scope.lower()
    if not utils.valid_string(args.type, min_length=2):
        raise ValueError(f"{args.type} is not a valid name")
    schema = args.type.lower()
    return path, name, scope, schema


def _build_structure(path, subdirs):
    for subdir in subdirs:
        os.makedirs(os.path.join(path, subdir), exist_ok=True)


def get_info(src, extracted=False):
    """Get information about a package.

    Using the package.json file located within the packages directory,
    fetch the information about the package including the version number
    and identifier.
    """
    if extracted:
        data = json.load(src)
    else:
        package = os.path.join(src, "package.json")
        with open(package) as package_data:
            data = json.load(package_data)
    return data



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Run package command tests."""
import argparse

from a28.package import generate_jsondata
from tests.conftest import Helpers


def test_a28_package_no_param(helpers: Helpers) -> None:
    """Ensure all subcommands are present."""
    command = ["package"]
    expect = "usage: a28 package [-h] {init,meta,build,install,publish} ..."

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0 : len(expect)] == expect


def test_a28_init_no_param(helpers: Helpers) -> None:
    """Test package initialization without parameters."""
    command = ["package", "init"]
    expect = "usage: a28 package init [-h] -s SCOPE -n NAME -t"

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0 : len(expect)] == expect


def test_a28_build_no_param(helpers: Helpers) -> None:
    """Test package build without parameters."""
    command = ["package", "build"]
    expect = "usage: a28 package build [-h] --src SRC [--dest DEST]"

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0 : len(expect)] == expect


def test_a28_install_no_param(helpers: Helpers) -> None:
    """Test package install without parameters."""
    command = ["package", "install"]
    expect = "usage: a28 package install [-h] --pkg PKG"

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0 : len(expect)] == expect


def test_generate_jsondata_without_args():
    """Check the json data generated is correct."""
    result = generate_jsondata(
        scope="foo",
        name="bar",
        identifier="1234",
        schema="234",
        args=argparse.Namespace(bin="", script=""),
    )

    assert result == {
        "description": "234 package created using A28 cmd by foo.",
        "identifier": "1234",
        "name": "@foo/bar",
        "version": "0.1.0",
    }


def test_generate_jsondata_with_bin():
    """Check the json data generated is correct with binary dir."""
    result = generate_jsondata(
        scope="foo",
        name="bar",
        identifier="1234",
        schema="234",
        args=argparse.Namespace(bin="in the bin", script=""),
    )

    assert result == {
        "description": "234 package created using A28 cmd by foo.",
        "identifier": "1234",
        "name": "@foo/bar",
        "version": "0.1.0",
        "bin": {},
    }


def test_generate_jsondata_with_bin_and_script():
    """Check the json data generated is correct with script and bin."""
    result = generate_jsondata(
        scope="foo",
        name="bar",
        identifier="1234",
        schema="234",
        args=argparse.Namespace(bin="in the bin", script="the script"),
    )

    assert result == {
        "description": "234 package created using A28 cmd by foo.",
        "identifier": "1234",
        "name": "@foo/bar",
        "version": "0.1.0",
        "bin": {},
        "scripts": {},
    }



"""Manage publication of packages."""
import logging
from argparse import Namespace, _SubParsersAction
from typing import Union

from a28 import utils
from a28.api import API, ApiError, JwtAuth
from a28.build import extract_meta
from a28.config import ConfigFile

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
log = logging.getLogger(__name__)


def add_publish_parser(parser: _SubParsersAction):
    """Create the CLI parser for 'publish' command."""
    parser.set_defaults(func=publish)
    parser.add_argument(
        "--pkg",
        required=True,
        help="package a28 package file",
    )
    parser.add_argument(
        "-e",
        "--endpoint",
        default=API.DEFAULT_ENDPOINT,
        help="endpoint to use.",
    )


def publish(args: Namespace):
    """Call the API to publish the version."""
    publish_version(args.endpoint, args.pkg)


def publish_version(region: str, path: str) -> None:
    """Publish a new version of the given a28 package file.

    1- We'll parse the .a28 zip file to retrieve package id/version.
    2- Generate a S3 signed URL.
    3- Upload the .a28 file on S3.
    4- Then create the version using the API.
    """
    log.debug(f'publishing a version: {{"region": {region}, "path": {path}}}')

    utils.message("Uploading", end="", flush=True)

    api = API(region)

    try:
        token = ConfigFile.get_token(api.region)
        meta_data = extract_meta(path)
        utils.message(".", end="", flush=True)
        file_name = _upload_a28_file(api, token, path)
        utils.message(".", end="", flush=True)
        api.post(
            f'/package/{meta_data["identifier"]}/version',
            json_payload={"package": file_name, "version": meta_data["version"]},
            auth=JwtAuth(token),
        )
        utils.message(".", end="", flush=True)
    except Exception as error:
        utils.message(f"\nFailed: {error}")
        raise

    utils.message("\nDone!")


def _upload_a28_file(api: API, token: str, path: str) -> Union[None, str]:
    signed_url_response = api.post(
        "/asset/generate",
        json_payload={
            "extension": "a28",
        },
        auth=JwtAuth(token),
    )

    with open(path, "rb") as package:
        upload_response = api.session.put(
            url=signed_url_response["data"]["url"], data=package
        )

    if upload_response.status_code != 200:
        raise ApiError(upload_response.status_code)

    return signed_url_response["data"]["name"]



"""Create build of a package."""
import json
import os
import zipfile
from argparse import _SubParsersAction

from a28 import utils

PKG_FILE = "package.json"


def add_build_parser(parser: _SubParsersAction):
    """Create the CLI parser for 'build' command."""
    parser.set_defaults(func=build)
    parser.add_argument(
        "--src",
        required=True,
        help="package source directory",
    )
    parser.add_argument(
        "--dest",
        default="",
        help="destination directory",
    )


def build(args):
    """Build package."""
    src = args.src
    dest = args.dest
    meta = get_info(src)
    pkg = package(src, dest, meta)
    utils.message(pkg)


def package(src, dest, meta):
    """Build an a28 package from the provided src directory.

    The package will be saved to the dest directory. A package needs to
    be provided containing at least an identifier and a version number.
    """
    version = meta["version"]
    identifier = meta["identifier"]
    filename = f"{identifier}-{version}.a28"
    filename = os.path.join(dest, filename)

    os.makedirs(dest, exist_ok=True)

    a28 = zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED)
    exclude = ["build", ".vscode", ".git", ".github", "dist"]
    for root, dirs, files in os.walk(src):
        dirs[:] = [d for d in dirs if d not in exclude]
        for current in files:
            i_file = os.path.join(root, current)
            fl = os.path.relpath(
                os.path.join(root, current),
                os.path.join(src),
            )
            a28.write(i_file, fl)
    a28.close()
    return filename


def get_info(src, extracted=False):
    """Get information about a package.

    Using the package.json file located within the packages directory,
    fetch the information about the package including the version number
    and identifier.
    """
    if extracted:
        data = json.load(src)
    else:
        package_config = os.path.join(src, PKG_FILE)
        with open(package_config) as package_data:
            data = json.load(package_data)
    return data


def extract_meta(pkg):
    """Extract the meta information from a packaged package."""
    with zipfile.ZipFile(pkg) as zf:
        package_config = next(x for x in zf.namelist() if x.endswith(PKG_FILE))
        with zf.open(package_config) as index:
            return get_info(index, True)



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Test system commands."""
from pathlib import Path

from tests.conftest import Helpers


def test_a28_system_no_param(helpers: Helpers) -> None:
    """Test execution of system sub argument without parameters."""
    command = ["system"]
    expect = "usage: a28 system [-h] (--environment {prod,staging,dev,sit}"

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0 : len(expect)] == expect


def test_a28_exists_no_param_no_config(helpers: Helpers, tmp_path: Path):
    """Test execution of system exists without config."""
    command = ["system", "-a", "exists"]
    config_path = tmp_path / "config_path"
    expect = f'No configuration exists at "{config_path}".'
    output, raise_type, exit_code = helpers.execute(command, config_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect == output


def test_a28_exists_no_param_config(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument without parameters."""
    command = ["system", "-a", "exists"]
    config_path = tmp_path / "config_path"
    expect = f'Configuration exists at "{config_path}".'
    config_path.mkdir()
    output, raise_type, exit_code = helpers.execute(command, config_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect == output


def test_a28_clean_no_param_no_config(helpers: Helpers, tmp_path: Path):
    """Test execution of system clean without config."""
    command = ["system", "-a", "clean", "-f"]
    config_path = tmp_path / "config_path"
    expect = f'No configuration to clean at "{config_path}".'
    output, raise_type, exit_code = helpers.execute(command, config_path)

    assert raise_type == SystemExit
    assert exit_code == 1
    assert expect == output


def test_a28_clean_no_param_config(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument without parameters."""
    command = ["system", "-a", "clean", "-f"]
    config_path = tmp_path / "config_path"
    expect = f"Deleted all files and directories in {config_path}"
    config_path.mkdir()
    output, raise_type, exit_code = helpers.execute(command, config_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect == output


def test_a28_path_minimal_dev(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument without parameters."""
    command = ["system", "-e=dev", "path", "--minimal"]
    expect = tmp_path / "dev"

    output, raise_type, exit_code = helpers.execute(command, tmp_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect.resolve() == Path(output).resolve()


def test_a28_path_minimal_prod(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument without parameters."""
    command = ["system", "-e=prod", "path", "--minimal"]
    expect = tmp_path / "prod"

    output, raise_type, exit_code = helpers.execute(command, tmp_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect.resolve() == Path(output).resolve()


def test_a28_path_minimal_all(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument with all parameter."""
    command = ["system", "--all", "path", "--minimal"]
    expect = tmp_path

    output, raise_type, exit_code = helpers.execute(command, tmp_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect.resolve() == Path(output).resolve()


def test_a28_path_no_minimal(helpers: Helpers, tmp_path: Path) -> None:
    """Test execution of system sub argument without minimal."""
    command = ["system", "--all", "path"]
    expect = f'Configuration path set to "{tmp_path.resolve()}".'
    output, raise_type, exit_code = helpers.execute(command, tmp_path)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert expect == output



"""Manage the CLI config file."""
import json
import os
from typing import Any, Dict

from a28 import utils


class ConfigError(Exception):
    """Custom exception for config errors."""

    pass


class ConfigFile:
    """Handles the config file storing token, endpoints and configs.

    `_save` is 'private' so that we're force to use ConfigFile as a
    context manager to load/save.
    """

    def __enter__(self) -> Dict[str, Any]:
        """Load the config file data on context creation."""
        self.data = self.load()
        return self.data

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Save the file to the disk on context exit."""
        return self._save(self.data)

    @staticmethod
    def get_token(region: str) -> str:
        """Retrieve the token from config.

        Load configuration information and return the token for the
        given region.
        """
        config = ConfigFile.load()

        if region not in config or "token" not in config[region]:
            raise ConfigError("please authenticate")

        return config[region]["token"]

    @staticmethod
    def load() -> Dict[str, Any]:
        """Load configuration information if information exists."""
        try:
            with open(utils.CONFIG) as json_file:
                data = json.load(json_file)
            return data
        except OSError:
            return {}
        except Exception as err:
            raise ConfigError(err)

    @staticmethod
    def _save(data: Dict[str, Any]) -> None:
        """Save all configuration values to the config file."""
        content = json.dumps(data, indent=4)

        # create the config path if it doesn't exist
        os.makedirs(utils.CONFIG_PATH, exist_ok=True)

        # save the configuration data to file
        with open(utils.CONFIG, "w") as config:
            config.write(content)
